@@ -8,7 +8,10 @@ import scipy
 import pandas as pd
 
 import cobra
-from cobra import Metabolite, Reaction, Gene
+#from cobra import Metabolite, Reaction, Gene
+from cobra import Gene
+from .metabolite import Metabolite
+from .reaction import Reaction
 from cobra.flux_analysis import deletion, moma, phenotype_phase_plane
 from cobra.core.solution import get_solution
 #from cobra.manipulation import modify
@@ -23,7 +26,10 @@ from ..io import Network
 
 
 class model(cobra.Model):
-    def __init__(self, existing_model=None, bounds=1000.0):
+    def __init__(self, existing_model=None, bounds=float('inf')):
+        self.all_reactions = {}
+        self.unusable_reactions = {}
+        self.all_mets = {}
         if type(existing_model) == model:
             self.__dict__ = existing_model.__dict__
         else:
@@ -33,6 +39,7 @@ class model(cobra.Model):
             self.quadratic_component = None
             self.bounds = bounds
             self.SetBounds(bounds=bounds)
+        
 
     #### MANIPULATING AND WRITING MODELS #########################################
 
@@ -56,6 +63,8 @@ class model(cobra.Model):
                 reac.id += sf
             for met in sf_model.metabolites:
                 met.id += sf
+                if isinstance(met.compartment, str): 
+                    met.compartment += sf
             sf_model.repair()
             big_model.MergeWithModel(sf_model)
         return big_model
@@ -74,7 +83,7 @@ class model(cobra.Model):
 #                   sbml_level=2, sbml_version=1, fbc=False, ExtReacs=[]):
 #        """ model_format = "sbml" | "excel" | "matlab" | "json" | "cobra" | "cobra_old" | "scrumpy" """
     def WriteModel(self, filename, model_format=None, excel_format="cobra", ExtReacs=[], **kwargs):
-        """ model_format = "sbml" | "sbml_legacy" | "excel" | "matlab" | "json" | "cobra" | "cobra_old" | "scrumpy" | "yaml" """
+        """ model_format = "sbml" | "sbml_legacy" | "excel" | "matlab" | "json" | "cobra" | "cobra_old" | "scrumpy" | "yaml" "cyc" """
         if self.id == None:
             self.id = 'None'
         from ..io import IO
@@ -215,17 +224,25 @@ class model(cobra.Model):
         self.add_reaction(reaction)
 
     #BUGGED DOES NOT DELETE THE REACTANT ADDED--> CAUSE BUG WITH: DEADENDMETABOLITES, PERIPHERALMETABOLITES():"Produced","Consumed",""
-    def DelReaction(self, reaction, delete_metabolites=False):
+    def DelReaction(self, reaction, delete_metabolites=False, clean=True):
         reaction = self.GetReaction(reaction)
         if delete_metabolites:
             for met in reaction.metabolites: 
-                self.DelMetabolite(met)
+                self.DelMetabolite(met,clean=clean)
         self.remove_reactions([reaction])
 
-    def DelReactions(self, reactions, delete_metabolites=False):
+        if clean:
+            if reaction.id in self.all_reactions:
+                del self.all_reactions[reaction.id]
+            if reaction.id in self.unusable_reactions:
+                del self.unusable_reactions[reaction.id]
+            if reaction.id in self.ReactionsWithNoFormulaMetabolites():
+                del self.reactions_mets_nf[reaction.id]
+
+    def DelReactions(self, reactions, delete_metabolites=False, clean=True):
         """ reactions = list of reactions """
         for reac in reactions:
-            self.DelReaction(reac, delete_metabolites=delete_metabolites)
+            self.DelReaction(reac, delete_metabolites=delete_metabolites, clean=clean)
         #self.remove_reactions(reactions)
 
     def ChangeReactionStoichiometry(self, reaction, metstoidic, combine=False):
@@ -322,8 +339,9 @@ class model(cobra.Model):
         else:
             metabolite = Metabolite(id=met, formula=formula, name=name,
                                 charge=charge,compartment=compartment)
+            self.add_metabolites([metabolite])
 
-    def DelMetabolite(self, met, destructive=False, method='substractive'):
+    def DelMetabolite(self, met, destructive=False, method='substractive',clean=True):
         """ method = 'subtractive'|'destructive' """
         met = self.GetMetabolite(met)
         if method == 'substractive': 
@@ -332,12 +350,27 @@ class model(cobra.Model):
         #    for reac in list(met._reaction):
                 #reac.remove_from_model()
         #        self.DelReaction(reac)
-        #met.remove_from_model(method=method)   
+        #met.remove_from_model(method=method)
+        l_reactions = list(met._reaction)
         met.remove_from_model(destructive=destructive)
+        if clean:
+            if self.all_mets is not None:
+                if met.id in self.all_mets:
+                    del self.all_mets[met.id]
+                if met.id in self.NoFormulaMetabolites():
+                    del self.no_formula_mets[met.id]
+            if destructive:
+                for reaction in l_reactions:
+                    if reaction.id in self.all_reactions:
+                        del self.all_reactions[reaction.id]
+                    if reaction.id in self.unusable_reactions:
+                        del self.unusable_reactions[reaction.id]
+                    if reaction.id in self.ReactionsWithNoFormulaMetabolites():
+                        del self.reactions_mets_nf[reaction.id]
 
-    def DelMetabolites(self, mets, method='subtractive'):
+    def DelMetabolites(self, mets, method='subtractive', clean=True):
         for met in mets:
-            self.DelMetabolite(met, method=method)
+            self.DelMetabolite(met, method=method, clean=clean)
 
     def SubstituteMetabolite(self, met_from, met_to):
         met_from = self.GetMetabolite(met_from)
@@ -392,6 +425,60 @@ class model(cobra.Model):
                                 neutral_formula += "H"
                 met.neutral_formula = neutral_formula
 
+    ######## SPECIALIZED METABOLITES AND REACTIONS METHODS: ADDED WITH CYC SUPPORT #########
+    def NoFormulaMetabolites(self, update=False):
+        """ get metabolites in model without molecular formula """
+        if not update and hasattr(self, 'no_formula_mets'):
+            return list(self.no_formula_mets.keys())
+
+        no_formula_mets = {}
+        result = []
+        for k in self.Metabolites():
+            if self.GetMetabolite(k).formula is None:
+                no_formula_mets[k] = self.GetMetabolite(k)
+                result.append(k)
+                
+        self.no_formula_mets = no_formula_mets
+        return result
+
+    def ReactionsWithNoFormulaMetabolites(self, update=False):
+        if not update and hasattr(self, 'reactions_mets_nf'):
+            return list(self.reactions_mets_nf.keys())
+
+        reactions_mets_nf = {}
+        mets = self.NoFormulaMetabolites()
+        vals = list(self.no_formula_mets.values())
+        for met in vals:
+            for r in list(met._reaction):
+                reactions_mets_nf[r.id] = r
+        self.reactions_mets_nf = reactions_mets_nf
+        return list(reactions_mets_nf.keys())
+
+    def RemoveNoFormulaMetabolites(self, exclude=[], with_reactions=True, destructive=True, method="destructive", clean=True):
+        """ exclude takes in a list of metabolites id: str """
+        l_mets = None
+        if hasattr(self, 'no_formula_mets'):
+            l_mets = list(self.no_formula_mets.keys())
+        else:
+            l_mets = self.NoFormulaMetabolites()
+        for k in l_mets:
+            if k in exclude:
+                continue
+            del self.no_formula_mets[k]
+            del self.all_mets[k]
+            self.DelMetabolite(k, destructive=destructive, method=method, clean=clean)
+        """  
+        if with_reactions:
+            collate = self.ReactionsWithNoForumlaMetabolites()
+            for k in reactions_mets_nf:
+                del reactions_mets_nf[k]
+                del all_reactions[k]
+                if k in unusable_reactions:
+                    del unusable_reactions[k]
+                self.DelReaction(k, delete_metabolites=delete_metabolites)
+        """
+        return
+
     ######## GETTING GENES ############################################
     def GetGene(self, gene):
         if not isinstance(gene, Gene):
@@ -452,31 +539,49 @@ class model(cobra.Model):
         #return double_deletion(self, element_list_1=element_list_1, element_list_2=element_list_2, method=method, single_deletion_growth_dict=single_deletion_growth_dict, element_type=element_type, solver=solver, number_of_processes=number_of_processes, return_frame=return_frame, zero_cutoff=zero_cutoff, **kwargs)
 
     #### CLEANING FUNCTIONS ########################################
-    def Clean(self,reac=True,met=True,gene=True):
-        if(met):
-            in_use_mets=self.CollateActiveMetabolites();
-            dead_end_mets = self.DeadEndMetabolites();
-            delete_ = []
-            for m in dead_end_mets:
-                if in_use_mets.get(m) is None:
-                    delete_.append(m)
+    #TODO: UPDATE CLEANING FUNCTION FOR DIFFERENT VARIABLES FOR CYC SUPPORT
+    def Clean(self,reac=True,met=True,gene=True,**kwargs):
+        if reac:
+            rs = self.Reactions()
+            for r in rs:
+                if len(self.GetReaction(r)._metabolites) == 0:
+                    self.DelReaction(r)
+            unused = False
+            if "unusable_reactions" in kwargs and kwargs["unusable_reactions"]:
+                unused = True
+            temp = self.all_reactions
+            for v in temp:
+                if v not in rs:
+                    del self.all_reactions[v]
+                    if unused and v in self.unusable_reactions:
+                        del self.unusable_reactions[v]
+                        if v in self.reactions_mets_nf:
+                            del self.reactions_mets_nf[v]
+        if met:
+            m_list = self.Metabolites()
+            for m in m_list:
+                if len(self.GetMetabolite(m)._reaction) == 0:
+                    self.DelMetabolite(m)
+
+    def TruncateCompartment(self, component):
+        """ component: reaction | metabolite """
+        truncated = []
+        
+        if component == "reaction":
+            component = "Reactions"
+        elif component == "metabolite":
+            component = "Metabolites"
+        else:
+            raise Exception("Unknown component: "+component)
+            
+        for v in getattr(self, component)():
+            a = v.split("_")
+            if len(a) > 1:
+                truncated.append(v[0:len(v)-len(a[len(a)-1])-1])
+            else:
+                truncated.append(v)
+        return truncated
             
-            self.DelMetabolites(delete_)
-
-
-    def CollateActiveMetabolites(self, reactions=None, AsMetNames=False):
-        in_use_mets={}
-        if reactions == None:
-            reactions = self.reactions
-        elif isinstance(reactions,str):
-            reactions = self.Reactions(reactions)
-        for reac in reactions:
-            r = self.GetReaction(reac)
-            for m in r.reactants:
-                in_use_mets[m._id]=1
-            for m in r.products:
-                in_use_mets[m._id]=1
-        return in_use_mets
 
     #### ASSOCIATIONS BETWEEN ATTRIBUTES #####################################
     
@@ -540,73 +645,6 @@ class model(cobra.Model):
             for elm in input_dict:
                 result[elm.id] = elm
             return result
-    ## SCAN FUNCTIONS ##################################################################
-    def ScanFBA(self, range_list, rxn_constrained, obj_direct, obj_dict, save_as=None):
-        """
-        range_list: list of values the constraint is scanned over, e.g. [0.1, 0.2, 0.3, 0.4]
-        rxn_constrained: name of reaction being constrained e.g.'CO2_tx1'
-        obj_direct: direction of the objective function, either "Max" or "Min"
-        obj_dict: dictionary of objective function, e.g. {'Photon_tx1':1,'Photon_tx2':1}
-        save_as: path to csv file, e.g. "/home/scobra/FBA_results.csv"
-
-        return value: a pandas dataframe containing the output data of the scan
-        """
-
-        df = pd.DataFrame(self.Reactions(), columns=["Reactions"])
-
-        for i in range_list:
-            self.SetConstraints({rxn_constrained:(i,i)})
-            self.SetObjDirec(obj_direct)
-            self.SetObjective(obj_dict)
-            try:
-                self.MinFluxSolve()
-            except Exception as e:
-                print("Error in parameter: " + str(i))
-                continue
-            sol=self.GetSol(AsMtx=True)
-            dic = {"Reactions": sol.index, ("Flux" + str(i)): list(sol[sol.columns[0]])}
-            flux_df = pd.DataFrame(dic)
-
-            df = pd.merge(df, flux_df, on=["Reactions"], how="outer")
-
-        if save_as:
-            df.to_csv(save_as)
-
-        return(df)
-
-    def ScanFVA(self, range_list, rxn_constrained, obj_direct, obj_dict, save_as=None):
-        """
-        range_list: list of values the constraint is scanned over, e.g. [0.1, 0.2, 0.3, 0.4]
-        rxn_constrained: name of reaction being constrained e.g.'CO2_tx1'
-        obj_direct: direction of the objective function, either "Max" or "Min"
-        obj_dict: dictionary of objective function, e.g. {'Photon_tx1':1,'Photon_tx2':1}
-        save_as: path to csv file, e.g. "/home/scobra/FBA_results.csv"
-
-        return value: a pandas dataframe containing the output data of the scan
-        """
-        
-        df = pd.DataFrame(self.Reactions(), columns=["Reactions"])
-
-        for i in range_list:
-            self.SetConstraints({rxn_constrained:(i,i)})
-            self.SetObjDirec(obj_direct)
-            self.SetObjective(obj_dict)
-            try:
-                sol=self.FVA(cobra=True)
-            except Exception as e:
-                print("Error in parameter: " + str(i))
-                continue
-
-            dic = {"Reactions": sol.keys(),
-                    ("Flux" + str(i)+"Value1"): [x[0] for x in sol.values()],
-                    ("Flux" + str(i)+"Value2"): [x[1] for x in sol.values()]}
-            flux_df = pd.DataFrame(dic)
-            df = pd.merge(df, flux_df, on=["Reactions"], how="outer")
-
-        if save_as:
-            df.to_csv(save_as)
-
-        return(df)
 
     @property
     def ReactionsToGenesAssociations(self):
@@ -683,7 +721,7 @@ class model(cobra.Model):
 
     ###### SETTING BOUNDS ###############################
 
-    def SetBounds(self, bounds=1000.0, thres=1000.0):
+    def SetBounds(self, bounds=float('inf'), thres=1000.0):
         for reac in self.reactions:
             if reac.lower_bound <= -thres or reac.lower_bound==None:
                 reac.lower_bound = -bounds
@@ -1020,12 +1058,12 @@ class model(cobra.Model):
         if not sol:
             #sol = flux(self.solution.x_dict
             #        ) if dict(self.solution.x_dict) != None else flux()
-            if self.solution.status == 'optimal':
+            if self.solution and self.solution.status == 'optimal':
                 #sol_object = Reversible.MergeSolution(self.solution)
                 #sol = flux(sol_object.fluxes.to_dict())
                 sol = flux(self.solution.fluxes.to_dict())
             else: 
-                print("no solution found")
+                print("no optimal solution")
                 sol = flux({})  
                 return sol
         else:
@@ -1211,6 +1249,71 @@ class model(cobra.Model):
 
 
     ### SCANS ###############################################################
+    def ScanFBA(self, range_list, rxn_constrained, obj_direct, obj_dict, save_as=None):
+        """
+        range_list: list of values the constraint is scanned over, e.g. [0.1, 0.2, 0.3, 0.4]
+        rxn_constrained: name of reaction being constrained e.g.'CO2_tx1'
+        obj_direct: direction of the objective function, either "Max" or "Min"
+        obj_dict: dictionary of objective function, e.g. {'Photon_tx1':1,'Photon_tx2':1}
+        save_as: path to csv file, e.g. "/home/scobra/FBA_results.csv"
+        return value: a pandas dataframe containing the output data of the scan
+        """
+
+        df = pd.DataFrame(self.Reactions(), columns=["Reactions"])
+
+        for i in range_list:
+            self.SetConstraints({rxn_constrained:(i,i)})
+            self.SetObjDirec(obj_direct)
+            self.SetObjective(obj_dict)
+            try:
+                self.MinFluxSolve()
+            except Exception as e:
+                print("Error in parameter: " + str(i))
+                continue
+            sol=self.GetSol(AsMtx=True)
+            dic = {"Reactions": sol.index, ("Flux" + str(i)): list(sol[sol.columns[0]])}
+            flux_df = pd.DataFrame(dic)
+
+            df = pd.merge(df, flux_df, on=["Reactions"], how="outer")
+
+        if save_as:
+            df.to_csv(save_as)
+
+        return(df)
+
+    def ScanFVA(self, range_list, rxn_constrained, obj_direct, obj_dict, save_as=None):
+        """
+        range_list: list of values the constraint is scanned over, e.g. [0.1, 0.2, 0.3, 0.4]
+        rxn_constrained: name of reaction being constrained e.g.'CO2_tx1'
+        obj_direct: direction of the objective function, either "Max" or "Min"
+        obj_dict: dictionary of objective function, e.g. {'Photon_tx1':1,'Photon_tx2':1}
+        save_as: path to csv file, e.g. "/home/scobra/FBA_results.csv"
+        return value: a pandas dataframe containing the output data of the scan
+        """
+
+        df = pd.DataFrame(self.Reactions(), columns=["Reactions"])
+
+        for i in range_list:
+            self.SetConstraints({rxn_constrained:(i,i)})
+            self.SetObjDirec(obj_direct)
+            self.SetObjective(obj_dict)
+            try:
+                sol=self.FVA(cobra=True)
+            except Exception as e:
+                print("Error in parameter: " + str(i))
+                continue
+
+            dic = {"Reactions": sol.keys(),
+                    ("Flux" + str(i)+"Value1"): [x[0] for x in sol.values()],
+                    ("Flux" + str(i)+"Value2"): [x[1] for x in sol.values()]}
+            flux_df = pd.DataFrame(dic)
+            df = pd.merge(df, flux_df, on=["Reactions"], how="outer")
+
+        if save_as:
+            df.to_csv(save_as)
+
+        return(df)
+
     def ConstraintScan(self, cd, lo, hi, n_p, MinFlux=True, IncZeroes=True,cobra=True):
         """ scan one reaction flux
             pre: cd = sum of reaction fluxes dictionary """
@@ -1410,3 +1513,4 @@ class model(cobra.Model):
     
         return comparison 
     
+
